-module(icfp).
-include("icfp.hrl").
-compile(export_all).

start([IPaddr, Port]) ->
	IP = atom_to_list(IPaddr),
	P = list_to_integer(atom_to_list(Port)),
	{ok, Socket} = gen_tcp:connect("127.0.0.1", 20005, [binary, {active, false}, {packet, raw}]),
%	{ok, Socket} = gen_tcp:connect(IP, P, [binary, {active, false}, {packet, raw}]),
	Atom = 'Player\n',
	S = list_to_binary(atom_to_list(Atom)),
	Res = gen_tcp:send(Socket, atom_to_list(Atom)),
	gen_tcp:controlling_process(Socket, self()),
	recv_init(Socket).
%
recv_init(Socket) ->
	BI = #bi{},
	case gen_tcp:recv(Socket, 0) of
	{ok, Bin} ->
		parse_binary(Socket, Bin, start, BI);
	X -> 
		exit(normal)
	end,
	recv(Socket).

recv(S) -> exit(normal).

parse_binary(S, <<>>, start, B) -> {done, B};
parse_binary(S, <<10:8, Rest/binary>>, start, B) when size(Rest) == 0 ->
	{done, B};
parse_binary(S, <<10:8, Rest/binary>>, start, B) ->
	parse_binary(S, Rest, get_row, B);
parse_binary(S, Bin, start, B) ->
	case get_dimensions(Bin, B) of
		{ok, <<>>, NewB} -> 
			BI = NewB,
			Rest = get_binary(S);
		{ok, Rest, NewB} -> BI = NewB
	end,
	parse_binary(S, Rest, get_row, BI);

parse_binary(S, Bin, get_row, BI) ->
	ets:new(base_table, [set, public, named_table, {keypos, 1}]),
	case get_rows(Bin, BI) of
	{done, <<>>, NewB} ->
		{ok, NextBin} = get_binary(S),
		parse_binary(S, NextBin, get_player, NewB);
	{done, Bin1, NewB} ->
		update_digraph(BI#bi.graph, BI),
		parse_binary(S, Bin1, get_player, NewB)
	end;

parse_binary(S, Bin, get_player, BI) ->
	ets:new(bot_table, [set, public, named_table, {keypos, 1}]),
	ets:new(pkg_table, [set, public, named_table, {keypos, 1}]),
	case get_player(Bin, BI) of
	{done, <<>>, NewB} ->
		{ok, NextBin} = get_binary(S),
		parse_binary(S, NextBin, get_init, NewB);
	{done, Bin1, NewB} ->
		parse_binary(S, Bin1, get_init, NewB)
	end;

parse_binary(S, Bin, get_init, BI) ->
	case get_init(Bin, BI) of
	{done, <<>>, NewB} ->
		{ok, NextBin} = get_binary(S),
		parse_binary(S, NextBin, get_packages, NewB);
	{done, Bin1, NewB} ->
		parse_binary(S, Bin1, get_packages, NewB)
	end;

parse_binary(S, Bin, get_packages, BI) ->
	play(S, suck_space(Bin), BI).

play(S, <<>>, BI) ->
	{ok, NewBin} = get_binary(S),
	play(S, NewBin, BI);

play(S, Bin, BI) ->
	ets:delete_all_objects(pkg_table),
	{MyX, MyY} = BI#bi.mypos,
%	L = ets:match(pkg_table, {'$2', MyX, MyY, '$3', '$4', '$1'}),
	B = ets:match(base_table, {'_','$1','$2'}),
	Rest = get_pkg_list(suck_space(Bin), BI),
	case should_i_drop(BI) of
		{none, _} -> ok;
		{drop, List} when list(List) ->
			{ok, DNewBI} = make_move(S, BI, drop, List),
			 {ok, BinD, MDNewBI} = get_moves(S, DNewBI),
			 play(S, BinD, MDNewBI);
		DE -> io:format("Error: should_i_drop gave ~w~n", [DE])
	end,
	case should_i_pick(BI) of
		{none, _} -> ok;
		{pick, PickList} when list(PickList) ->
			{ok, PNewBI} = make_move(S, BI, pick, PickList),
			{ok, BinP, MPNewBI} = get_moves(S, PNewBI),
			play(S, BinP, MPNewBI);
		PE -> io:format("Error: should_i_pick gave ~w~n", [PE])
	end,

	case should_i_move(BI) of
		{none, _} -> ok;
		{Direction, Dist} ->
			{ok, MNewBI} = make_move(S, BI, move, Direction),
			{ok, BinM, MMNewBI} = get_moves(S, MNewBI),
			play(S, BinM, MMNewBI);
		ME -> io:format("Error: should_i_move gave ~w~n", [ME])
	end,

	case should_i_get_to_base(BI) of
		{none, _} -> ok;
		{BDirection, BDist} ->
			{ok, BNewBI} = make_move(S, BI, move, BDirection),
			{ok, BinB, MBNewBI} = get_moves(S, BNewBI),
			play(S, BinB, MBNewBI);
		BE -> io:format("Error: should_i_get_to_base gave ~w~n", [BE])
	end.

get_binary(S) -> 
	case gen_tcp:recv(S, 0) of
	{ok, Bin} -> {ok, Bin};
	X -> {ok , <<>>}
end.

get_dimensions(Bin, B) ->
	{Y, Rest} = get_int(Bin),
	{X, Rest1} = get_int(Rest),
	NewB = B#bi{rows=X, cols=Y, bdim= done},
%	{ok, Rest1, B#bi{rows=X, cols=Y, bdim= done}}.
	init_graph(NewB#bi.graph, 1, 1, X, Y),
	<<10:8, Rest2/binary>> = Rest1,
	{ok, Rest2, NewB}.

get_int(<<32:8, Rest/binary>>) -> get_int(Rest);
get_int(<<>>) -> error;
get_int(B) -> get_int(B, 0).

get_int(<<C:8,Rest/binary>>, Acc) when $0 =< C, C =< $9 ->
	get_int(Rest, 10*Acc + C - $0);
get_int(Bin, Acc) -> {Acc, Bin}.

init_graph(G, X, Y, X, Y) -> 
	digraph:add_vertex(G, {X,Y});
init_graph(G, NowX, Y, X, Y) ->
	digraph:add_vertex(G, {NowX, Y}),
	init_graph(G,NowX+1, 1, X, Y);
init_graph(G, X, NowY, X, Y) ->
	digraph:add_vertex(G, {X, NowY}),
	init_graph(G,X, NowY+1, X, Y);
init_graph(G, NowX, NowY, X, Y) ->
	digraph:add_vertex(G, {NowX, NowY}),
	init_graph(G, NowX, NowY+1, X, Y).

% get_rows(Bin, BI) -> {done, NewBin or <<>>, NewB}.

get_rows(<<10:8, Rest/binary>>, BI) ->
	Row = BI#bi.rows_compl,
	X = BI#bi.rows,
	case X-1 of
	Row -> 
		{done, Rest, BI#bi{rows_compl=X, board=done}};
	_Y ->
		get_rows(Rest, BI#bi{rows_compl=Row+1, cols_compl=0})
	end;

get_rows(<<>>, BI) ->
	{done, <<>>, BI};

get_rows(<<46:8,Rest/binary>>, BI) ->
	Row = BI#bi.rows_compl,
	Col = BI#bi.cols_compl,
	digraph:add_vertex(BI#bi.graph, {Row+1, Col+1}, [safe]),
	get_rows(Rest, BI#bi{cols_compl = Col+1});

get_rows(<<35:8,Rest/binary>>, BI) ->
	Row = BI#bi.rows_compl,
	Col = BI#bi.cols_compl,
	digraph:add_vertex(BI#bi.graph, {Row+1, Col+1}, [wall]),
	get_rows(Rest, BI#bi{cols_compl = Col+1});

get_rows(<<64:8,Rest/binary>>, BI) ->
	Row = BI#bi.rows_compl,
	Col = BI#bi.cols_compl,
	digraph:add_vertex(BI#bi.graph, {Row+1, Col+1}, [home_base]),
	ets:insert(base_table, [{now(), Row+1, Col+1}]),
	get_rows(Rest, BI#bi{cols_compl = Col+1});

get_rows(<<126:8,Rest/binary>>, BI) ->
	Row = BI#bi.rows_compl,
	Col = BI#bi.cols_compl,
	digraph:add_vertex(BI#bi.graph, {Row+1, Col+1}, [lethal]),
	get_rows(Rest, BI#bi{cols_compl = Col+1});

get_rows(<<C:8, Bin/binary>>, BI) -> 
	io:format("error unknown char ~w~n", [C]).

update_digraph(G, BI) ->
	X = BI#bi.rows,
	Y = BI#bi.cols,
	update_digraph(G, 1, 1, X, Y).

% update_digraph(G, LocX, LocY, MaxX, MaxY) -> returns nothing

update_digraph(G, X, Y, X, Y) -> 
	check_vertex(G, X, Y, X, Y);

update_digraph(G, NowX, Y, X, Y) ->
	check_vertex(G, NowX, Y, X, Y),
	update_digraph(G, NowX+1, 1, X, Y);

update_digraph(G, X, NowY, X, Y) ->
	check_vertex(G, X, NowY, X, Y),
	update_digraph(G, X, NowY+1, X, Y);
	
update_digraph(G, NowX, NowY, X, Y) ->
	check_vertex(G, NowX, NowY, X, Y),
	update_digraph(G, NowX, NowY+1, X, Y).

check_vertex(G, NowX, NowY, X, Y) ->
	{V, [Label]} = digraph:vertex(G, {NowX, NowY}),
	if 
		Label == safe ->
			update_vertex(G, NowX, NowY, X, Y);
		Label == home_base ->
			update_vertex(G, NowX, NowY, X, Y);
		true -> skip
	end.

update_vertex(G, 1, 1, X, Y) ->
	update_north(G, 1, 1),
	update_east(G, 1, 1);

update_vertex(G, X, Y, X, Y) ->
	update_south(G, X, Y),
	update_west(G, X, Y);

update_vertex(G, 1, Y, X, Y) ->
	update_north(G, 1, Y),
	update_west(G, 1, Y);

update_vertex(G, X, 1, X, Y) ->
	update_south(G, X, 1),
	update_east(G, X, 1);

update_vertex(G, 1, NowY, X, Y) ->
	update_north(G, 1, NowY),
	update_east(G, 1, NowY),
	update_west(G, 1, NowY);

update_vertex(G, X, NowY, X, Y) ->
	update_south(G, X, NowY),
	update_east(G, X, NowY),
	update_west(G, X, NowY);

update_vertex(G, NowX, 1, X, Y) ->
	update_north(G, NowX, 1),
	update_south(G, NowX, 1),
	update_east(G, NowX, 1);

update_vertex(G, NowX, Y, X, Y) ->
	update_north(G, NowX, Y),
	update_south(G, NowX, Y),
	update_west(G, NowX, Y);

update_vertex(G, NowX, NowY, X, Y) ->
	update_north(G, NowX, NowY),
	update_south(G, NowX, NowY),
	update_east(G, NowX, NowY),
	update_west(G, NowX, NowY).

update_east(G, X, Y) ->
	{V, [Label]} = digraph:vertex(G, {X, Y+1}),
	if 
		Label == safe ->
			digraph:add_edge(G, {X, Y}, {X, Y+1});
		Label == home_base ->
			digraph:add_edge(G, {X, Y}, {X, Y+1});
	true -> 
		skip
	end.

update_west(G, X, Y) ->
	{V, [Label]} = digraph:vertex(G, {X, Y-1}),
	if 
		Label == safe ->
			digraph:add_edge(G, {X, Y}, {X, Y-1});
		Label == home_base ->
			digraph:add_edge(G, {X, Y}, {X, Y-1});
	true ->
		skip
	end.

update_north(G, X, Y) ->
	{V, [Label]} = digraph:vertex(G, {X+1, Y}),
	if 
		Label == safe ->
			digraph:add_edge(G, {X, Y}, {X+1, Y});
		Label == home_base ->
			digraph:add_edge(G, {X, Y}, {X+1, Y});
	true ->
		skip
	end.

update_south(G, X, Y) ->
	{V, [Label]} = digraph:vertex(G, {X-1, Y}),
	if 
		Label == safe ->
			digraph:add_edge(G, {X, Y}, {X-1, Y});
		Label == home_base ->
			digraph:add_edge(G, {X, Y}, {X-1, Y});
	true ->
		skip
	end.


get_player(Bin, BI) ->
	{Id, Rest} = get_int(Bin),
	{MaxCap, Rest1} = get_int(Rest),
	{Money, Rest2} = get_int(Rest1),
	<<10:8,Rest3/binary>> = Rest2,
	{done , Rest3, BI#bi{mylimit=MaxCap, myid=Id, mymoney=Money, config=done}}.

% get_init(Bin, Bi) -> {done, Bin, BI}

get_init(<<>>, BI) ->
	error;
get_init(<<10:8, Rest/binary>>, BI) ->
	{done, Rest, BI#bi{init=done}};

get_init(<<35:8, Rest/binary>>, BI) ->
	{Id, Bin1} = get_int(Rest),
	<<88:8, Rest1/binary>> = suck_space(Bin1),
	{X, Bin2} = get_int(Rest1),
	<<89:8, Rest2/binary>> = suck_space(Bin2),
	{Y, Bin3} = get_int(Rest2),
	ets:insert(bot_table, {Id, X, Y}),
	if 
		BI#bi.myid == Id ->
			NewB = BI#bi{mypos={X,Y}};
		true ->
			NewB = BI
	end,
	get_init(suck_space(Bin3), NewB).

suck_space(<<>>) -> <<>>;
suck_space(<<32:8, Rest/binary>>) -> suck_space(Rest);
suck_space(Bin) -> Bin.

%get_pkg_list(Bin, BI) -> Bin (updates pkg_list in ets)

get_pkg_list(<<10:32, Rest/binary>>, BI) -> 
	Rest;

get_pkg_list(<<10:8>>, BI) -> 
	<<>>;

get_pkg_list(<<>>, BI) -> 
	<<>>;

get_pkg_list(Bin, BI) ->
	{PId, Bin1} = get_int(Bin),
	{X, Bin2} = get_int(suck_space(Bin1)),
	{Y, Bin3} = get_int(suck_space(Bin2)),
	{Wt, Bin4} = get_int(suck_space(Bin3)),
	{MyX, MyY} = BI#bi.mypos,
	ets:insert(pkg_table, {PId, MyX, MyY, X, Y, Wt}),
	get_pkg_list(suck_space(Bin4), BI).

get_move(BI) ->
	PkgList = ets:match(pkg_table, '$1'),
	{Pkg, Val} = pick_pkg(BI, PkgList),
	{Pkg, Val}.

pick_pkg(BI, List) -> 
	pick_pkg(BI, List, current_weight(BI#bi.pkglist), {none, 0}).

pick_pkg(BI, [], CWT, M) -> M;
pick_pkg(BI, [[PId, MyX, MyY, X, Y, Wt]|T], CWT, {Id, Val}) ->
	MyWt = BI#bi.mylimit,
	if 
	CWT < MyWt ->
		if
		CWT + Wt < MyWt ->
			Value = heur_val(Wt, 
			length(digraph:get_short_path(BI#bi.graph,
			{MyX, MyY}, {X, Y}))),
			if 
				Value > Val ->
				pick_pkg(BI, T, CWT, {PId, Value});
				true ->
				pick_pkg(BI, T, CWT, {Id, Val})
			end;
		true ->
			pick_pkg(BI, T, CWT, {Id, Val})
		end;			
	true -> {none, 0}
	end.
	

current_weight(L) -> 
	current_weight(L, 0).

current_weight([], Sum) -> Sum;
current_weight([{PId, X, Y, Wt}|T], Sum) ->
	current_weight(T, Sum+Wt).

heur_val(W, D) when integer(W), integer(D) -> 45*W/(D+W);
heur_val(W, D) -> 0.

should_i_pick(BI) ->
	{MyX, MyY} = BI#bi.mypos,
	MyLimit = BI#bi.mylimit,
	MyPkgs = BI#bi.pkglist,
	CWT=current_weight(MyPkgs),
	FullList = ets:match(pkg_table, '$1'),
	AvalilPkgs = ets:match(pkg_table, {'$2', MyX, MyY, '$3', '$4', '$1'}),
	SortedList = lists:reverse(lists:sort(AvalilPkgs)),
	should_i_pick(SortedList, CWT, CWT, [], MyLimit, MyX, MyY).

should_i_pick([], CWT, EWT, [], MyLimit, MyX, MyY) -> {none, 0};
should_i_pick([], CWT, EWT, [], MyLimit, MyX, MyY) -> {none, 0};
should_i_pick([], CWT, EWT, List, MyLimit, MyX, MyY) -> {pick, List};
should_i_pick([[Wt, Pid, X, Y]|T], CWT, EWT, List, MyLimit, MyX, MyY) when MyLimit =< EWT + Wt  ->
	should_i_pick(T, CWT, EWT, List, MyLimit, MyX, MyY);
should_i_pick([[Wt, Pid, X, Y]|T], CWT, EWT, List, MyLimit, MyX, MyY) ->
	should_i_pick(T, CWT, EWT+Wt, [{Wt, Pid, X, Y}|List], MyLimit, MyX, MyY).

should_i_drop(BI) ->
	{MyX, MyY} = BI#bi.mypos,
	MyPkgs = BI#bi.pkglist,
	should_i_drop(MyPkgs, MyX, MyY, []).

should_i_drop([], MyX, MyY, []) -> {none, 0};
should_i_drop([], MyX, MyY, []) -> {none, 0};
should_i_drop([], MyX, MyY, List) -> {drop, List};
should_i_drop([{Pid, MyX, MyY, Wt}|T], MyX, MyY, List) ->
	should_i_drop(T, MyX, MyY, [Pid|List]);
should_i_drop([H|T], MyX, MyY, List) ->
	should_i_drop(T, MyX, MyY, List).

should_i_move(BI) ->
	{MyX, MyY} = BI#bi.mypos,
	MyPkgs = BI#bi.pkglist,
	should_i_move(BI#bi.graph, MyPkgs, MyX, MyY, {none,?LARGE_NUMBER}).

should_i_move(G, [], MyX, MyY, Move) -> Move;
should_i_move(G, [{Pid, X, Y, Wt}|T], MyX, MyY, {Dir, Val}) ->
	Path = digraph:get_short_path(G, {MyX, MyY}, {X, Y}),
	Dist = length(Path),
	if 
		Dist =< Val ->
			NextStep = lists:nth(2, Path),
			{ok, Direction} = get_direction({MyX, MyY}, NextStep),
			should_i_move(G, T, MyX, MyY, {Direction, Dist});
		true ->
			should_i_move(G, T, MyX, MyY, {Dir, Val})
	end.
	

should_i_get_to_base(BI) ->
	{MyX, MyY} = BI#bi.mypos,
	BaseList = ets:match(base_table, {'_','$1','$2'}),
	should_i_get_to_base(BI, BI#bi.graph, BaseList, MyX, MyY, {none, ?LARGE_NUMBER}).

should_i_get_to_base(BI, G, [], MyX, MyY, {Dir, Val}) -> {Dir, Val};
should_i_get_to_base(BI, G, [[X,Y]|T], MyX, MyY, {Dir, Val}) ->
	OldPos = BI#bi.oldpos,
	Path = digraph:get_short_path(G, {MyX, MyY}, {X, Y}),
	Dist = length(Path) - 1,
	if 
	Dist =< Val ->
		NextStep = lists:nth(2, Path),
		if 
		OldPos == NextStep ->
			should_i_get_to_base(BI, G, T, MyX, MyY, {Dir, Val});
		true ->
			{ok, Direction} = get_direction({MyX, MyY}, NextStep),
			should_i_get_to_base(BI, G, T, MyX, MyY, {Direction, Dist})
		end;
	true ->
			should_i_get_to_base(BI, G, T, MyX, MyY, {Dir, Val})
	end.

get_direction({MyX, MyY}, {X, Y}) when X==MyX, Y==MyY+1 -> {ok, east};
get_direction({MyX, MyY}, {X, Y}) when X==MyX, Y==MyY-1 -> {ok, west};
get_direction({MyX, MyY}, {X, Y}) when X==MyX+1, Y==MyY -> {ok, north};
get_direction({MyX, MyY}, {X, Y}) when X==MyX-1, Y==MyY -> {ok, south};

get_direction(A, B) -> 
	error.

%make_move(S, BI, pick/drop/move, List/List/Dir) -> {ok, New BI}

make_move(S, BI, pick, List) when list(List) ->
	do_pick(S, BI, List);
make_move(S, BI, pick, List) ->
	error;

make_move(S, BI, drop, List) when list(List) ->
	do_drop(S, BI, List);
make_move(S, BI, drop, List) ->
	error;

make_move(S, BI, move, D) when D == east; D == west; D == north; D == south
	-> do_move(S, BI, D);
make_move(S, BI, move, D) ->
	error.

do_pick(S, BI, List) ->
	Money=BI#bi.mymoney,
	Bin = get_pick_bin(BI, List),
	Bin1 = concat_binary([<<49:8,32:8,80:8,105:8,99:8,107:8,32:8>>, Bin, <<10:8>>]),
	ok = gen_tcp:send(S, Bin1),
	{ok, BI#bi{mymoney=Money-1}}.

do_drop(S, BI, List) ->
	Money=BI#bi.mymoney,
	Bin = get_drop_bin(BI, List),
	Bin1 = concat_binary([<<49:8,32:8,68:8,114:8,111:8,112:8,32:8>>, Bin, <<10:8>>]),
%	{ok, NewBI} = clean(BI, List),
	ok = gen_tcp:send(S, Bin1),
	{ok, BI#bi{mymoney=Money-1}}.

do_move(S, BI, D) ->
	ok = gen_tcp:send(S, concat_binary([<<49:8>>, get_dir_bin(D),<<10:8>>])),
	Money = BI#bi.mymoney,
	{ok, BI#bi{mymoney=Money-1}}.

get_dir_bin(east) -> <<32:8,77:8,111:8,118:8,101:8,32:8,69:8>>;
get_dir_bin(west) -> <<32:8,77:8,111:8,118:8,101:8,32:8,87:8>>;
get_dir_bin(north) -> <<32:8,77:8,111:8,118:8,101:8,32:8,78:8>>;
get_dir_bin(south) -> <<32:8,77:8,111:8,118:8,101:8,32:8,83:8>>.

%get_moves(S, BI) -> {ok, Bin, NewBI}.

get_moves(S, BI) ->
	case gen_tcp:recv(S, 0) of
	{error, Error} -> 
			  erlang:halt();
	{ok, Bin} ->
		  parse_board_moves(Bin, BI)
	end.


%parse_board_moves(Bin, BI) -> {ok, NewBin, NewBI}.

parse_board_moves(<<>>, BI) -> error;
parse_board_moves(<<10:8, Rest/binary>>, BI) -> {ok, Rest, BI};
parse_board_moves(<<35:8, Rest/binary>>, BI) ->
		{NewBin, NewBI} = get_one_move(suck_space(Rest), BI),
		parse_board_moves(NewBin, NewBI);
parse_board_moves(Bin, BI) -> error.

%get_one_move(Bin, BI) -> {NewBin, NewBI} (NewBin starts with #

get_one_move(Bin, BI) ->
	{Id, NewBin} =  get_int(Bin),
	{MoveList, New1} = get_actions(suck_space(NewBin), Id),
	{ok , NewBI} = handle_actions(MoveList, Id, BI),
	{New1, NewBI}.

%get_actions(Bin) -> Bin has to start with E/S/W/N/X/D/P. Should return
% complete move (as a list) and a bin starting with a #

get_actions(Bin, Id) -> get_actions(Bin, Id, #move{id=Id, actions=[]}).

get_actions(<<10:8,Rest/binary>>, Id, Moves) -> 
	L = Moves#move.actions,
	{lists:reverse(L), concat_binary([<<10:8>>, Rest])};

get_actions(<<88:8, Rest/binary>>, Id, Moves) ->
	List = Moves#move.actions,
	{X, Bin} = get_int(suck_space(Rest)),
	<<32:8,89:8,Rest1/binary>> = Bin,
	{Y, Bin1} = get_int(suck_space(Rest1)),
	get_actions(Bin1, Id, Moves#move{appeared=yes, 
				actions = [ {appeared, X, Y} | List]});

get_actions(<<32:8, Rest/binary>>, Id, M) ->
	get_actions(Rest, Id, M);

get_actions(<<69:8, Rest/binary>>, Id, M) ->
	L = M#move.actions,
	get_actions(Rest, Id, M#move{actions=[{move, east}|L]});

get_actions(<<83:8, Rest/binary>>, Id, M) ->
	L = M#move.actions,
	get_actions(Rest, Id, M#move{actions=[{move, south}|L]});

get_actions(<<87:8, Rest/binary>>, Id, M) ->
	L = M#move.actions,
	get_actions(Rest, Id, M#move{actions=[{move, west}|L]});

get_actions(<<78:8, Rest/binary>>, Id, M) ->
	L = M#move.actions,
	get_actions(Rest, Id, M#move{actions=[{move, north}|L]});

get_actions(<<68:8, Rest/binary>>, Id, M) ->
	L = M#move.actions,
	{PId, Bin} = get_int(Rest),
	get_actions(Bin, Id, M#move{actions=[{drop, PId}|L]});

get_actions(<<80:8, Rest/binary>>, Id, M) ->
	L = M#move.actions,
	{PId, Bin} = get_int(Rest),
	get_actions(Bin, Id, M#move{actions=[{pick, PId}|L]});

get_actions(<<35:8,Rest/binary>>, Id, M) ->
	L = M#move.actions,
	{concat_binary([<<35:8>>,Rest]), M#move{actions=lists:reverse(L)}};

get_actions(Bin, Id, M) -> error.

%handle_actions(Move list, Id, BI) -> {ok, NewBI}

handle_actions([], Id, BI) -> {ok, BI};
handle_actions(List, Id, #bi{myid=MId} = BI) when MId == Id ->
	handle_my_actions(List, BI);

handle_actions([{appeared, X, Y} | T], Id, BI) ->
	G = BI#bi.graph,
	ets:insert(bot_table, {Id, X, Y}),
	{V, Label} = digraph:vertex(G, {X, Y}),
	digraph:add_vertex(G, {X,Y}, lists:append(Label, [Id])),
	handle_actions(T, Id, BI);

handle_actions([{drop, P}|T], Id, BI) ->
	handle_actions(T, Id, BI);

handle_actions([{pick, P}|T], Id, BI) ->
	handle_actions(T, Id, BI);

handle_actions([{move, Dir}|T], Id, BI) when Dir == east; Dir == west; Dir == north; Dir == south ->
	[{Bid, X, Y}] = ets:lookup(bot_table, Id),
	{NewX, NewY} = update_move_in_digraph(BI#bi.graph, Id, {X, Y}, Dir),
	ets:insert(bot_table, {Id, NewX, NewY}),
	handle_actions(T, Id, BI);

handle_actions([{move, Dir}|T], Id, BI) -> 
	handle_actions(T, Id, BI).

update_move_in_digraph(G, Id, {OldX, OldY}, Dir) ->
	{V, Label} = digraph:vertex(G, {OldX, OldY}),
	NewX = OldX + xincr(Dir),
	NewY = OldY + yincr(Dir),
	{V1, L1} = digraph:vertex(G, {NewX, NewY}),
	OldList = lists:delete(Id, Label),
	NewList = lists:append(L1, [Id]),
	digraph:add_vertex(G, {OldX, OldY}, OldList),
	digraph:add_vertex(G, {NewX, NewY}, NewList),
	{NewX, NewY}.

xincr(east) -> 0;
xincr(west) -> 0;
xincr(north) -> 1;
xincr(south) -> -1.

yincr(east) -> 1;
yincr(west) -> -1;
yincr(north) -> 0;
yincr(south) -> 0.

handle_my_actions([{appeared, X, Y} | T], BI) ->
	handle_my_actions(T, BI);

handle_my_actions([], BI) -> {ok, BI};
handle_my_actions([{pick, Id} | T], BI) ->
	L = BI#bi.pkglist,
	[{PId, MyX, MyY, X, Y, Wt}] = ets:lookup(pkg_table, Id),
	handle_my_actions(T, BI#bi{pkglist=lists:append(L, [{PId, X, Y, Wt}])});

handle_my_actions([{drop, Id}|T], BI) ->
	L = BI#bi.pkglist,
	{PId, X, Y, Wt} = find_id_in_list(Id, L),
	handle_my_actions(T, BI#bi{pkglist=lists:delete({PId, X, Y, Wt}, L)});

handle_my_actions([{move, Dir} | T], BI) when Dir == east; Dir == west; Dir == north; Dir == south ->
	Id = BI#bi.myid,
	OldPos = BI#bi.mypos,
	[{Bid, X, Y}] = ets:lookup(bot_table, Id),
	{NewX, NewY} = update_move_in_digraph(BI#bi.graph, Id, {X, Y}, Dir),
	ets:insert(bot_table, {Id, NewX, NewY}),
	handle_my_actions(T, BI#bi{mypos={NewX, NewY}, oldpos=OldPos});

handle_my_actions([{move, Dir} | T], BI) -> 
	handle_my_actions(T, BI);

handle_my_actions([H|T], BI) -> io:format("Wrong elem ~w~n", [H]), error.


find_id_in_list(Id, []) -> error;
find_id_in_list(Id, [{Id, X, Y, W}|T]) -> {Id, X, Y, W};
find_id_in_list(Id, [H|T]) -> find_id_in_list(Id, T).

get_pick_bin(BI, List) -> get_pick_bin(BI, List, []).

get_pick_bin(BI, [], Acc) -> concat_binary(lists:reverse(Acc));
get_pick_bin(BI, [{Wt, Id, X, Y}|T], Acc) ->
	Bin = concat_binary([list_to_binary(integer_to_list(Id)), <<32:8>>]),
	get_pick_bin(BI, T, [Bin|Acc]).

get_drop_bin(BI, List) -> get_drop_bin(BI, List, []).

get_drop_bin(BI, [], Acc) -> concat_binary(lists:reverse(Acc));
get_drop_bin(BI, [H|T], Acc) ->
	Bin = concat_binary([list_to_binary(integer_to_list(H)), <<32:8>>]),
	get_drop_bin(BI, T, [Bin|Acc]).

clean(BI, List) ->
	OldList=BI#bi.pkglist,
	clean(BI, List, OldList).

clean(BI, [], OldList) -> {ok, BI#bi{pkglist=OldList}};
clean(BI, [H|T], OldList) ->
	clean(BI, T, get_new_list(H, OldList)).

get_new_list(Elem, List) -> get_new_list(Elem, List, []).

get_new_list(Elem, [], Acc) -> Acc;
get_new_list(Elem, [{Elem, X, Y, Wt}|T], Acc) -> lists:append(Acc, T);
get_new_list(Elem, [H|T], Acc) ->
	get_new_list(Elem, T, [H|Acc]).

new_get_rows(Bin, BI) ->
	new_get_rows(BI, BI#bi.graph, Bin, 1, 1, BI#bi.rows, BI#bi.cols+1).

new_get_rows(BI, G, <<C:8,Rest/binary>>, X, Y, X, Y) ->
	{done, Rest, BI#bi{rows_compl=X, cols_compl=Y-1}};

new_get_rows(BI, G, <<C:8,Rest/binary>>, CurX, Y, X, Y) ->
	new_get_rows(BI, G, Rest, CurX+1, 1, X, Y);

new_get_rows(BI, G, <<C:8, Rest/binary>>, CurX, CurY, X, Y) ->
	init_digraph(G, C, CurX, CurY),
	new_get_rows(BI, G, Rest, CurX, CurY+1, X, Y).

init_digraph(G, 46, X, Y) ->
	digraph:add_vertex(G, {X, Y}, [safe]);

init_digraph(G, 35, X, Y) ->
	digraph:add_vertex(G, {X, Y}, [wall]);

init_digraph(G, 126, X, Y) ->
	digraph:add_vertex(G, {X, Y}, [lethal]);

init_digraph(G, 64, X, Y) ->
	digraph:add_vertex(G, {X, Y}, [home_base]);

%init_digraph(G, 10, X, Y) -> ok;

init_digraph(G, C, X, Y) -> ok. % io:format("Recd ~w at ~w ~w~n", [C, X, Y]).
